#include "main.h"
#include <getopt.h>
#include "hrd.h"
#include "mica.h"

FILE* log_fp = stdout;

#include "../../sRDMA/utilities/cxxopts.hpp"

cxxopts::ParseResult parse(int argc, char* argv[]) {
  cxxopts::Options options(argv[0],
                           "Secure QP bandwidth test with multiple workers");
  options.positional_help("[optional args]").show_positional_help();

  try {
    std::stringstream stream;
    stream << "0x" << std::hex << IBV_HDR_HMAC_SHA2_256;
    std::string resultstr(stream.str());

    options.add_options()("smartnic", "smartnic address",
                          cxxopts::value<std::string>(), "IP")(
        "threads", "Total number of QP created",
        cxxopts::value<uint32_t>()->default_value(std::to_string(1)),
        "N")("drkey", "Use key derivation",
             cxxopts::value<bool>()->default_value("false"))(
        "client", "client", cxxopts::value<bool>()->default_value("false"))(
        "updatepercentage", "updatepercentage",
        cxxopts::value<int>()->default_value(std::to_string(0)))(
        "crypto", "Protection algorithm",
        cxxopts::value<uint32_t>()->default_value(resultstr),
        "N")("memkey", "Use key derivation for memory sub delegation",
             cxxopts::value<bool>()->default_value("false"))(
        "batch", "batch",
        cxxopts::value<uint32_t>()->default_value(std::to_string(32)),
        "N")("outstand", "Number of outstanding requests",
             cxxopts::value<uint32_t>()->default_value(std::to_string(96)),
             "N")("o,output", "Output file", cxxopts::value<std::string>(),
                  "FILE")("help", "Print help");

    auto result = options.parse(argc, argv);

    if (result.count("help")) {
      std::cout << options.help({""}) << std::endl;
      exit(0);
    }

    if (!result.count("smartnic")) {
      throw new cxxopts::OptionException("Must specify smartnic ip address");
    }

    return result;

  } catch (const cxxopts::OptionException& e) {
    std::cout << "error parsing options: " << e.what() << std::endl;
    std::cout << options.help({""}) << std::endl;
    exit(1);
  }
}

int main(int argc, char* argv[]) {
  /* Use small queues to reduce cache pressure */
  assert(HRD_Q_DEPTH == 128);

  /* All requests should fit into the master's request region */
  assert(sizeof(struct mica_op) * NUM_CLIENTS * NUM_WORKERS * WINDOW_SIZE <
         RR_SIZE);

  /* Unsignaled completion checks. worker.c does its own check w/ @postlist */
  assert(UNSIG_BATCH >= WINDOW_SIZE);     /* Pipelining check for clients */
  assert(HRD_Q_DEPTH >= 2 * UNSIG_BATCH); /* Queue capacity check */

  int i;
  int num_threads = -1;
  int update_percentage = -1;
  struct thread_params* param_arr;
  pthread_t* thread_arr;

  auto allparams = parse(argc, argv);

  bool drkey = allparams["drkey"].as<bool>();
  uint32_t crypto = allparams["crypto"].as<uint32_t>();
  bool withmemkey = allparams["memkey"].as<bool>();
  num_threads = allparams["threads"].as<uint32_t>();
  bool is_client = allparams["client"].as<bool>();
  update_percentage = allparams["updatepercentage"].as<int>();

  /* Handle the master process specially */

  std::string my_nick_address = allparams["smartnic"].as<std::string>();

  if (allparams.count("output")) {
    log_fp = fopen(allparams["output"].as<std::string>().c_str(), "w");
  } else {
    log_fp = stdout;
  }

  RDMA_COM* dma_host =
      new RDMA_COM(0,            // inline size of QP
                   HRD_Q_DEPTH,  // send_size
                   HRD_Q_DEPTH,  // recv_size
                   0,            // we don't init reads
                   10,           // but we  target for 20 reads
                   8888,
                   my_nick_address.c_str());  // my_nick_address.c_str()

  text(log_fp, " DMA connection is established\n");

  int num_conn_qps = is_client ? num_threads : NUM_CLIENTS;

  Connection** all_dma_connections =
      (Connection**)malloc(sizeof(Connection*) * num_conn_qps);

  for (int conn = 0; conn < num_conn_qps; conn++) {
    all_dma_connections[conn] = dma_host->get_connection(2);
  }

  struct hrd_ctrl_blk* connections = NULL;
  if (!is_client) {
    connections = run_master(all_dma_connections, crypto, drkey, withmemkey);
  }

  /* Common sanity checks for worker process and per-machine client process */

  if (is_client) {
    assert(num_threads >= 1);
    assert(update_percentage >= 0 && update_percentage <= 100);
  } else {
    /* Server does not need to know number of client ports */
    num_threads = NUM_WORKERS; /* Needed to allocate thread structs later */
  }

  /* Launch a single server thread or multiple client threads */
  printf("main: Using %d threads\n", num_threads);
  param_arr =
      (struct thread_params*)malloc(num_threads * sizeof(struct thread_params));
  thread_arr = (pthread_t*)malloc(num_threads * sizeof(pthread_t));

  for (i = 0; i < num_threads; i++) {
    param_arr[i].crypto = crypto;
    param_arr[i].drkey = drkey;
    param_arr[i].withmemkey = withmemkey;
    if (is_client) {
      param_arr[i].seccon = &all_dma_connections[i];
      param_arr[i].id = i;
      param_arr[i].update_percentage = update_percentage;

      pthread_create(&thread_arr[i], NULL, run_client, &param_arr[i]);
    } else {
      param_arr[i].id = i;
      param_arr[i].connection = &connections[i];
      pthread_create(&thread_arr[i], NULL, run_worker, &param_arr[i]);
    }
  }

  for (i = 0; i < num_threads; i++) {
    pthread_join(thread_arr[i], NULL);
  }

  return 0;
}
