#pragma once
#include <stdint.h>

#include "hrd.h"
#include "mica.h"
/*
 * The polling logic in HERD requires the following:
 * 1. 0 < MICA_OP_GET < MICA_OP_PUT < HERD_OP_GET < HERD_OP_PUT
 * 2. HERD_OP_GET = MICA_OP_GET + HERD_MICA_OFFSET
 * 3. HERD_OP_PUT = MICA_OP_PUT + HERD_MICA_OFFSET
 *
 * This allows us to detect HERD requests by checking if the request region
 * opcode is more than MICA_OP_PUT. And then we can convert a HERD opcode to
 * a MICA opcode by subtracting HERD_MICA_OFFSET from it.
 */
#define HERD_MICA_OFFSET 10
#define HERD_OP_GET (MICA_OP_GET + HERD_MICA_OFFSET)
#define HERD_OP_PUT (MICA_OP_PUT + HERD_MICA_OFFSET)

#define HERD_NUM_BKTS (2 * 1024 * 1024)
#define HERD_LOG_CAP (1024 * 1024 * 1024)

#define HERD_NUM_KEYS (8 * 1024 * 1024)
#define HERD_VALUE_SIZE 32

/* Request sizes */
#define HERD_GET_REQ_SIZE (16 + 1) /* 16 byte key + opcode */

/* Key, op, len, val */
#define HERD_PUT_REQ_SIZE (16 + 1 + 1 + HERD_VALUE_SIZE)

/* Configuration options */
#define MAX_SERVER_PORTS 4
#define NUM_WORKERS 1
#define NUM_CLIENTS 8

/* Performance options */
#define WINDOW_SIZE 32 /* Outstanding requests kept by each client */
//#define NUM_UD_QPS 1   /* Number of UD QPs per port */
#define USE_POSTLIST 0

#define UNSIG_BATCH 64 /* XXX Check if increasing this helps */
#define UNSIG_BATCH_ (UNSIG_BATCH - 1)

/* SHM key for the 1st request region created by master. ++ for other RRs.*/
#define MASTER_SHM_KEY 24
#define RR_SIZE (16 * 1024 * 1024) /* Request region size */
#define OFFSET(wn, cn, ws) \
  ((wn * NUM_CLIENTS * WINDOW_SIZE) + (cn * WINDOW_SIZE) + ws)

struct thread_params {
  int id;
  int update_percentage;
  void* connection;
  void* seccon;
  uint32_t crypto;
  bool drkey;
  bool withmemkey;
};

struct hrd_ctrl_blk* run_master(void* cons, uint32_t crypto, bool drkey,
                                bool memkey);
void* run_worker(void* arg);
void* run_client(void* arg);

/* Random local_hids for master */
#define MASTER_P0_ID -22

struct hrd_ctrl_blk* run_master(void* rawcons, uint32_t crypto, bool drkey,
                                bool memkey) {
  int i, j;

  hrd_red_printf(
      "Running HERD master"
      "RR_SIZE = %d MB, RR use fraction = %.2f\n",
      RR_SIZE / M_1, (double)sizeof(struct mica_op) * NUM_CLIENTS *
                         NUM_WORKERS * WINDOW_SIZE / RR_SIZE);

  /*
   * Create one control block per port. Each control block has enough QPs
   * for all client QPs. The ith client QP must connect to the ith QP in a
   * a control block, but clients can choose the control block.
   */
  // struct hrd_ctrl_blk** cb = malloc(num_server_ports * sizeof(void*));
  struct hrd_ctrl_blk* cb;
  // assert(cb != NULL);

  /* Allocate a registered buffer for port #0 only */

  int shm_key = MASTER_SHM_KEY;

  Connection** cons = (Connection**)rawcons;

  cb = hrd_ctrl_blk_init(MASTER_P0_ID, /* local hid */
                         0,            /*  numa node id */
                         NUM_CLIENTS,  /* #conn_qps*/
                         NULL, RR_SIZE, shm_key, 0, 0,
                         -1, /* #dgram qps, buf size, shm key */
                         cons, crypto, drkey, memkey);

  /* Zero out the request buffer */
  memset((void*)cb->conn_buf, 0, RR_SIZE);

  /* Register all created QPs - only some will get used! */
  for (j = 0; j < NUM_CLIENTS; j++) {
    char srv_name[HRD_QP_NAME_SIZE];
    sprintf(srv_name, "master-0-%d", j);
    hrd_publish_conn_qp(cb, j, srv_name);
  }

  printf("main: Master published all QPs \n");

  hrd_red_printf("main: Master published all QPs. Waiting for clients..\n");

  for (i = 0; i < NUM_CLIENTS; i++) {
    char clt_conn_qp_name[HRD_QP_NAME_SIZE];
    sprintf(clt_conn_qp_name, "client-conn-%d", i);
    printf("main: Master waiting for client %s\n", clt_conn_qp_name);

    struct hrd_qp_attr* clt_qp = NULL;
    while (clt_qp == NULL) {
      clt_qp = hrd_get_published_qp(clt_conn_qp_name);
      if (clt_qp == NULL) {
        usleep(200000);
      }
    }

    printf("main: Master found client %d of %d clients. Connecting..\n", i,
           NUM_CLIENTS);

    hrd_connect_qp(cb, i, clt_qp);

    char mstr_qp_name[HRD_QP_NAME_SIZE];
    sprintf(mstr_qp_name, "master-0-%d", i);
    hrd_publish_ready(mstr_qp_name);
  }

  /*
   * Wait until the sun rises in the west and sets in the east. Until the
   * rivers run dry, and the mountains blow in the wind like leaves...
   */
  printf("main: Master leaving\n");
  // sleep(1000000);
  return cb;
}

void* run_worker(void* arg) {
  int ret;
  struct thread_params params = *(struct thread_params*)arg;
  int wrkr_lid = params.id; /* Local ID of this worker thread*/

  struct hrd_ctrl_blk* rccb = (struct hrd_ctrl_blk*)params.connection;

  assert(HERD_VALUE_SIZE <= MICA_MAX_VALUE);

  /* MICA instance id = wrkr_lid, NUMA node = 0 */
  struct mica_kv kv;
  mica_init(&kv, wrkr_lid, 0, HERD_NUM_BKTS, HERD_LOG_CAP);
  mica_populate_fixed_len(&kv, HERD_NUM_KEYS, HERD_VALUE_SIZE);

  /* Map the request region created by the master */
  volatile struct mica_op* req_buf;
  //  int sid = shmget(MASTER_SHM_KEY, RR_SIZE, SHM_HUGETLB | 0666);
  //  assert(sid != -1);

  req_buf = (volatile struct mica_op*)rccb->conn_buf;
  // assert(req_buf != (void*)-1);

  struct ibv_mr* mr =
      ibv_reg_mr(rccb->pd, kv.ht_log, HERD_LOG_CAP,
                 IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE |
                     IBV_ACCESS_REMOTE_READ);

  assert(mr != NULL);

  int ws[NUM_CLIENTS] = {0}; /* Per-client window slot */

  /* We can detect at most NUM_CLIENTS requests in each step */
  struct mica_op* op_ptr_arr[NUM_CLIENTS];
  struct mica_resp resp_arr[NUM_CLIENTS];
  // struct ibv_send_wr wr[NUM_CLIENTS]; //  TODO  , *bad_send_wr = NULL;
  struct ibv_sge sgl[NUM_CLIENTS];

  struct ibv_wc wc;
  long long rolling_iter = 0;         /* For throughput measurement */
  long long nb_tx[NUM_CLIENTS] = {0}; /* CQE polling */
                                      /* UD QP index: same for both ports */
  long long nb_tx_tot = 0;
  long long nb_post_send = 0;

  /*
   * @cb_i is the control block to use for @clt_i's response. If NUM_CLIENTS
   * is a multiple of @num_server_ports, we can cycle @cb_i in the client loop
   * to maintain cb_i = clt_i % num_server_ports.
   *
   * @wr_i is the work request to use for @clt_i's response. We use contiguous
   * work requests for the responses in a batch. This is because (in HERD) we
   * need to  pass a contiguous array of operations to MICA, and marshalling
   * and unmarshalling the contiguous array will be expensive.
   */

  printf("Done!");
  //  sleep(10000);

  int clt_i = 0;
  int poll_i, wr_i;

  int cltreq[NUM_CLIENTS];

  struct timespec start, end;
  clock_gettime(CLOCK_REALTIME, &start);

  while (1) {
    if (unlikely(rolling_iter >= M_4)) {
      clock_gettime(CLOCK_REALTIME, &end);
      double seconds = (end.tv_sec - start.tv_sec) +
                       (double)(end.tv_nsec - start.tv_nsec) / 1000000000;
      info(log_fp,
           "main: Worker %d: %.2f IOPS. Avg per-port postlist = %.2f. "
           "HERD lookup fail rate = %.4f\n",
           wrkr_lid, M_4 / seconds, (double)nb_tx_tot / nb_post_send,
           (double)kv.num_get_fail / kv.num_get_op);

      rolling_iter = 0;
      nb_tx_tot = 0;
      nb_post_send = 0;

      clock_gettime(CLOCK_REALTIME, &start);
    }

    /* Do a pass over requests from all clients */
    int nb_new_req = 0; /* New requests from port i*/
    wr_i = 0;

    for (poll_i = 0; poll_i < NUM_CLIENTS; poll_i++) {
      ret = rccb->secqp[poll_i]->poll_send_cq(&wc);

      /*
       * This cycling of @clt_i and @cb_i needs to be before polling. This
       * should be the only place where we modify @clt_i and @cb_i.
       */
      HRD_MOD_ADD(clt_i, NUM_CLIENTS);
      // assert(cb_i == clt_i % num_server_ports);	/* XXX */

      int req_offset = OFFSET(wrkr_lid, clt_i, ws[clt_i]);
      if (req_buf[req_offset].opcode < HERD_OP_GET) {
        continue;
      }

      /* Convert to a MICA opcode */
      req_buf[req_offset].opcode -= HERD_MICA_OFFSET;

      op_ptr_arr[wr_i] = (struct mica_op*)&req_buf[req_offset];
      cltreq[wr_i] = clt_i;

      HRD_MOD_ADD(ws[clt_i], WINDOW_SIZE);

      rolling_iter++;
      nb_tx[clt_i]++; /* Must increment inside loop */
      nb_tx_tot++;
      nb_new_req++;
      wr_i++;

      if (wr_i == NUM_CLIENTS) {
        break;
      }
    }

    if (wr_i) {
      mica_batch_op(&kv, wr_i, op_ptr_arr, resp_arr);

      int nb_new_req_tot = wr_i;
      for (wr_i = 0; wr_i < nb_new_req_tot; wr_i++) {
        sgl[wr_i].length = resp_arr[wr_i].val_len;
        sgl[wr_i].addr = (uint64_t)(uintptr_t)resp_arr[wr_i].val_ptr;
        sgl[wr_i].lkey = mr->lkey;
        bool signaled =
            true;  //(nb_tx[clt_i] % WINDOW_SIZE == 0) ? true : false;
        int ret = rccb->secqp[cltreq[wr_i]]->Send(0, &sgl[wr_i], signaled);
        if (ret) {
          printf("error on send [%d] %d\n", cltreq[wr_i], ret);
        }
        nb_post_send++;
      }
    }

    //    if(nb_post_send%100 == 0) printf("%u\n",nb_post_send);

    for (poll_i = 0; poll_i < NUM_CLIENTS; poll_i++) {
      ret = rccb->secqp[poll_i]->poll_send_cq(&wc);
      if (ret) {
      }
    }
  }

  return NULL;
}

#include "hrd.h"
#include "main.h"
#include "mica.h"

#define DGRAM_BUF_SIZE 4096

/* Generate a random permutation of [0, n - 1] for client @clt_gid */
int* get_random_permutation(int n, int clt_gid, uint64_t* seed) {
  int i, j, temp;
  assert(n > 0);

  /* Each client uses a different range in the cycle space of fastrand */
  for (i = 0; i < clt_gid * HERD_NUM_KEYS; i++) {
    hrd_fastrand(seed);
  }

  printf("client %d: creating a permutation of 0--%d. This takes time..\n",
         clt_gid, n - 1);

  int* log = (int*)malloc(n * sizeof(int));
  assert(log != NULL);
  for (i = 0; i < n; i++) {
    log[i] = i;
  }

  printf("\tclient %d: shuffling..\n", clt_gid);
  for (i = n - 1; i >= 1; i--) {
    j = hrd_fastrand(seed) % (i + 1);
    temp = log[i];
    log[i] = log[j];
    log[j] = temp;
  }
  printf("\tclient %d: done creating random permutation\n", clt_gid);

  return log;
}

void* run_client(void* arg) {
  int i;
  struct thread_params params = *(struct thread_params*)arg;
  int clt_gid = params.id; /* Global ID of this client thread */
  uint32_t update_percentage = (uint32_t)params.update_percentage;

  /* This is the only port used by this client */

  /*
   * TODO: The client creates a connected buffer because the libhrd API
   * requires a buffer when creating connected queue pairs. This should be
   * fixed in the API.
   */
  struct hrd_ctrl_blk* cb = hrd_ctrl_blk_init(
      clt_gid,               /* local_hid */
      -1,                    /* numa_node_id */
      1,                     /* #conn qps*/
      NULL, 4096, -1,        /* prealloc conn buf, buf size, key */
      1, DGRAM_BUF_SIZE, -1, /* num_dgram_qps, dgram_buf_size, key */
      (Connection**)params.seccon, params.crypto, params.drkey,
      params.withmemkey);

  char mstr_qp_name[HRD_QP_NAME_SIZE];
  sprintf(mstr_qp_name, "master-0-%d", clt_gid);

  char clt_conn_qp_name[HRD_QP_NAME_SIZE];
  sprintf(clt_conn_qp_name, "client-conn-%d", clt_gid);

  hrd_publish_conn_qp(cb, 0, clt_conn_qp_name);

  printf("main: Client %s published conn and dgram. Waiting for master %s\n",
         clt_conn_qp_name, mstr_qp_name);

  struct hrd_qp_attr* mstr_qp = NULL;
  while (mstr_qp == NULL) {
    mstr_qp = hrd_get_published_qp(mstr_qp_name);
    if (mstr_qp == NULL) {
      usleep(200000);
    }
  }

  printf("main: Client %s found master! Connecting..\n", clt_conn_qp_name);
  hrd_connect_qp(cb, 0, mstr_qp);
  hrd_wait_till_ready(mstr_qp_name);

  /* Start the real work */
  uint64_t seed = 0xdeadbeef;
  int* key_arr = get_random_permutation(HERD_NUM_KEYS, clt_gid, &seed);
  int key_i, ret;

  /* Some tracking info */
  int ws[NUM_WORKERS] = {0}; /* Window slot to use for a worker */

  const uint32_t TOTALBUF = 128;
  uint32_t currentbuf = 0;

  struct mica_op* req_buf =
      (struct mica_op*)memalign(4096, TOTALBUF * sizeof(*req_buf));
  assert(req_buf != NULL);

  struct ibv_mr* mr =
      ibv_reg_mr(cb->pd, req_buf, TOTALBUF * sizeof(*req_buf),
                 IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE |
                     IBV_ACCESS_REMOTE_READ);

  struct ibv_wc wc;

  long long rolling_iter = 0; /* For throughput measurement */
  long long nb_tx = 0;        /* Total requests performed or queued */
  int wn = 0;                 /* Worker number */

  struct timespec start, end;
  clock_gettime(CLOCK_REALTIME, &start);

  /* Fill the RECV queue */
  for (i = 0; i < HRD_Q_DEPTH; i++) {
    cb->secqp[0]->post_recv((uint64_t)(void*)cb->dgram_buf, DGRAM_BUF_SIZE,
                            cb->dgram_buf_mr->lkey, 0);
  }

  uint32_t pending = 0;

  const uint32_t batch = 16;

  printf("Done!");
  sleep(20);
  printf("\nStart!\n");

  while (1) {
    //    sleep(10000);
    if (rolling_iter >= K_512) {
      clock_gettime(CLOCK_REALTIME, &end);
      double seconds = (end.tv_sec - start.tv_sec) +
                       (double)(end.tv_nsec - start.tv_nsec) / 1000000000;
      info(log_fp, "main: Client %d: %.2f IOPS. nb_tx = %lld\n", clt_gid,
           K_512 / seconds, nb_tx);

      rolling_iter = 0;

      clock_gettime(CLOCK_REALTIME, &start);
    }

    if (pending > 0) {
      ret = cb->secqp[0]->poll_recv_cq(&wc);
      if (ret) {
        cb->secqp[0]->post_recv((uint64_t)(void*)cb->dgram_buf, DGRAM_BUF_SIZE,
                                cb->dgram_buf_mr->lkey, 0);

        if (wc.opcode == IBV_WC_SECURE_SEND) {
          pending--;
          // printf("[%d] Get reply\n",clt_gid);
        }
      }
    }

    if (pending < WINDOW_SIZE - 1) {
      wn = hrd_fastrand(&seed) % NUM_WORKERS; /* Choose a worker */
      int is_update = (hrd_fastrand(&seed) % 100 < update_percentage) ? 1 : 0;

      /* Forge the HERD request */
      key_i = hrd_fastrand(&seed) % HERD_NUM_KEYS; /* Choose a key */

      *(uint128*)&req_buf[currentbuf] = CityHash128((char*)&key_arr[key_i], 4);
      req_buf[currentbuf].opcode = is_update ? HERD_OP_PUT : HERD_OP_GET;
      req_buf[currentbuf].val_len = is_update ? HERD_VALUE_SIZE : -1;

      bool signaled = (nb_tx != 0 && nb_tx % batch == 0) ? true : false;

      //          printf("[ %d ] Send from qpn %p %u", clt_gid, cb->secqp[0],
      //          cb->secqp[0]->GetQPN());

      ret = cb->secqp[0]->Write(
          0, (uint64_t)(uintptr_t)&req_buf[currentbuf], mr->lkey,
          is_update ? HERD_PUT_REQ_SIZE : HERD_GET_REQ_SIZE,
          mstr_qp->buf_addr +
              OFFSET(wn, clt_gid, ws[wn]) * sizeof(struct mica_op),
          mstr_qp->rkey, signaled,
          mstr_qp->region.length == 0 ? NULL : &mstr_qp->region);

      HRD_MOD_ADD(currentbuf, TOTALBUF);
      CPE(ret, "ibv_post_send error", ret);
      pending++;

      // printf("Client %d: sending request index %lld\n", clt_gid, nb_tx);

      rolling_iter++;
      nb_tx++;
      HRD_MOD_ADD(ws[wn], WINDOW_SIZE);
    }

    ret = cb->secqp[0]->poll_send_cq(&wc);
    //    CPE(ret, "poll_send_cq error", ret);
  }

  return NULL;
}
