#ifndef HRD_H
#define HRD_H

#include <assert.h>
#include <errno.h>
#include <numaif.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <infiniband/verbs.h>
#include <libmemcached/memcached.h>
#include <malloc.h>
#include <time.h>
#include "hrd_sizes.h"

#include "../../sRDMA/worker/secure_qp.hpp"

/* Registry info about a QP */
struct hrd_qp_attr {
  char name[HRD_QP_NAME_SIZE];

  /* Info about the RDMA buffer associated with this QP */
  uintptr_t buf_addr;
  uint32_t buf_size;
  uint32_t rkey;

  uint32_t qpn;

  region_t region;
};

struct hrd_ctrl_blk {
  int local_hid;

  bool drkey;
  bool withmemkey;
  enum ibv_qp_crypto crypto;

  /* Info about the device/port to use for this control block */
  int numa_node_id; /* NUMA node id */

  struct ibv_pd* pd; /* A protection domain for this control block */
  SequreQP** secqp;
  /* Connected QPs */
  int num_conn_qps;

  volatile uint8_t* conn_buf; /* A buffer for RDMA over RC/UC QPs */
  int conn_buf_size;
  int conn_buf_shm_key;
  struct ibv_mr* conn_buf_mr;

  /* Datagram QPs */

  volatile uint8_t* dgram_buf; /* A buffer for RECVs on dgram QPs */
  int dgram_buf_size;
  int dgram_buf_shm_key;
  struct ibv_mr* dgram_buf_mr;
};

/* Major initialzation functions */
struct hrd_ctrl_blk* hrd_ctrl_blk_init(
    int local_hid, int numa_node_id, int num_conn_qps,

    volatile void* prealloc_conn_buf, int conn_buf_size, int conn_buf_shm_key,
    int num_dgram_qps, int dgram_buf_size, int dgram_buf_shm_key,
    Connection** cons, uint32_t crypto, bool drkey, bool withmemkey);

int hrd_ctrl_blk_destroy(struct hrd_ctrl_blk* cb);

void hrd_create_dgram_qps(struct hrd_ctrl_blk* cb);

void hrd_connect_qp(struct hrd_ctrl_blk* cb, int conn_qp_idx,
                    struct hrd_qp_attr* remote_qp_attr);

/* Fill @wc with @num_comps comps from this @cq. Exit on error. */
static inline void hrd_poll_cq(struct ibv_cq* cq, int num_comps,
                               struct ibv_wc* wc) {
  int comps = 0;

  while (comps < num_comps) {
    int new_comps = ibv_poll_cq(cq, num_comps - comps, &wc[comps]);
    if (new_comps != 0) {
      /* Ideally, we should check from comps -> new_comps - 1 */
      if (wc[comps].status != 0) {
        fprintf(stderr, "Bad wc status %d\n", wc[comps].status);
        exit(0);
      }

      comps += new_comps;
    }
  }
}

/* Registry functions */
void hrd_publish(const char* key, void* value, int len);
int hrd_get_published(const char* key, void** value);

/* Publish the nth connected queue pair from this cb with this name */
void hrd_publish_conn_qp(struct hrd_ctrl_blk* cb, int n, const char* qp_name);

struct hrd_qp_attr* hrd_get_published_qp(const char* qp_name);

void hrd_publish_ready(const char* qp_name);
void hrd_wait_till_ready(const char* qp_name);

/* Utility functions */
static inline uint32_t hrd_fastrand(uint64_t* seed) {
  *seed = *seed * 1103515245 + 12345;
  return (uint32_t)(*seed >> 32);
}

static inline long long hrd_get_cycles() {
  unsigned low, high;
  unsigned long long val;
  asm volatile("rdtsc" : "=a"(low), "=d"(high));
  val = high;
  val = (val << 32) | low;
  return val;
}

static inline int hrd_is_power_of_2(uint32_t n) { return n && !(n & (n - 1)); }

int hrd_free(int shm_key, void* shm_buf);
void hrd_red_printf(const char* format, ...);
void hrd_get_formatted_time(char* timebuf);
void hrd_nano_sleep(int ns);
char* hrd_getenv(const char* name);

unsigned char key[] = {0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6,
                       0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f, 0x3c,  // 16
                       0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6,
                       0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f, 0x3c};  // 32

unsigned char pdkey[] = {0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6,
                         0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f, 0x3c};  // 16

unsigned char mymemkey[] = {0x7e, 0x7e, 0x15, 0x16, 0x28, 0x7e,
                            0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88,
                            0x09, 0xcf, 0x4f, 0x3c};  // 16

/*
 * If @prealloc_conn_buf != NULL, @conn_buf_size is the size of the preallocated
 * buffer. If @prealloc_conn_buf == NULL, @conn_buf_size is the size of the
 * new buffer to create.
 */
struct hrd_ctrl_blk* hrd_ctrl_blk_init(
    int local_hid, int numa_node_id, /* -1 means don't use hugepages */
    int num_conn_qps, volatile void* prealloc_conn_buf, int conn_buf_size,
    int conn_buf_shm_key, int num_dgram_qps, int dgram_buf_size,
    int dgram_buf_shm_key, Connection** cons, uint32_t crypto, bool drkey,
    bool withmemkey) {
  hrd_red_printf(
      "HRD: creating control block %d:  socket %d, "
      "conn qps %d,  conn buf %d bytes (key %d), "
      "dgram qps %d, dgram buf %d bytes (key %d)\n",
      local_hid, numa_node_id, num_conn_qps, conn_buf_size, conn_buf_shm_key,
      num_dgram_qps, dgram_buf_size, dgram_buf_shm_key);

  /*
   * Check arguments for sanity.
   * @local_hid can be anything: it's just control block identifier that is
   * useful in printing debug info.
   */

  assert(numa_node_id >= -1 && numa_node_id <= 8);
  assert(num_conn_qps >= 0);
  assert(conn_buf_size >= 0 && conn_buf_size <= M_1024);

  /* If there is no preallocated buffer, shm key can be =/>/< 0 */
  if (prealloc_conn_buf != NULL) {
    assert(conn_buf_shm_key == -1);
  }
  assert(num_dgram_qps >= 0 && num_dgram_qps <= M_2);
  assert(dgram_buf_size >= 0 && dgram_buf_size <= M_1024);

  if (num_conn_qps == 0 && num_dgram_qps == 0) {
    hrd_red_printf(
        "HRD: Control block initialization without QPs. Are you"
        " sure you want to do this?\n");
    assert(false);
  }

  struct hrd_ctrl_blk* cb =
      (struct hrd_ctrl_blk*)malloc(sizeof(struct hrd_ctrl_blk));
  memset(cb, 0, sizeof(struct hrd_ctrl_blk));

  /* Fill in the control block */
  cb->local_hid = local_hid;
  cb->numa_node_id = numa_node_id;

  cb->num_conn_qps = num_conn_qps;
  cb->conn_buf_size = conn_buf_size;
  cb->conn_buf_shm_key = conn_buf_shm_key;

  cb->dgram_buf_size = dgram_buf_size;
  cb->dgram_buf_shm_key = dgram_buf_shm_key;

  cb->drkey = drkey;
  cb->withmemkey = withmemkey;
  cb->crypto = (enum ibv_qp_crypto)crypto;

  int ib_flags =
      IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_WRITE;

  /*
   * Create connected QPs and transition them to RTS.
   * Create and register connected QP RDMA buffer.
   */
  if (num_conn_qps >= 1) {
    cb->secqp = (SequreQP**)malloc(num_conn_qps * sizeof(SequreQP*));
    for (int conn = 0; conn < num_conn_qps; conn++) {
      text(log_fp, "[hrd_ctrl_blk_init] Create qp %d\n", conn);
      if (withmemkey) printf("With memkey;\n");
      cb->secqp[conn] = new SequreQP(cons[conn], drkey ? pdkey : NULL,
                                     withmemkey ? mymemkey : NULL, 4096 * 10);
    }

    cb->pd = cons[0]->id->qp->pd;

    if (conn_buf_size) {
      /* Create and register conn_buf - always make it multiple of 2 MB*/
      int reg_size = 0;

      /* If numa_node_id < 0, use standard heap memory */

      printf("reg1\n");
      reg_size = conn_buf_size;
      cb->conn_buf = (volatile uint8_t*)memalign(4096, reg_size);
      assert(cb->conn_buf != NULL);

      memset((char*)cb->conn_buf, 0, reg_size);
      cb->conn_buf_mr =
          cb->secqp[0]->reg_mem((void*)cb->conn_buf, reg_size, withmemkey);
      //   ibv_reg_mr(cb->pd, (char*)cb->conn_buf, reg_size, ib_flags);
      assert(cb->conn_buf_mr != NULL);
      if (withmemkey) {
        for (int conn = 1; conn < num_conn_qps; conn++) {
          cb->secqp[conn]->reg_mem(cb->conn_buf_mr);
        }
      }

    } else {
      cb->conn_buf_mr = NULL;
      cb->conn_buf = NULL;
    }
  }

  if (num_dgram_qps >= 1) {
    /* Create and register dgram_buf */
    int reg_size = 0;

    reg_size = dgram_buf_size;
    cb->dgram_buf = (volatile uint8_t*)memalign(4096, reg_size);

    assert(cb->dgram_buf != NULL);
    memset((char*)cb->dgram_buf, 0, reg_size);

    cb->dgram_buf_mr =
        ibv_reg_mr(cb->pd, (char*)cb->dgram_buf, reg_size, ib_flags);
    assert(cb->dgram_buf_mr != NULL);
  }

  return cb;
}

/* Free up the resources taken by @cb. Return -1 if something fails, else 0. */
int hrd_ctrl_blk_destroy(struct hrd_ctrl_blk* cb) {
  hrd_red_printf("HRD: Destroying control block %d\n", cb->local_hid);

  if (cb->num_conn_qps > 0) {
    free((void*)cb->conn_buf);
  }

  hrd_red_printf("HRD: Control block %d destroyed.\n", cb->local_hid);
  return 0;
}

/* Connects @cb's queue pair index @n to remote QP @remote_qp_attr */
void hrd_connect_qp(struct hrd_ctrl_blk* cb, int n,
                    struct hrd_qp_attr* remote_qp_attr) {
  text(log_fp, "[hrd_connect_qp] Connect %d\n", n);
  if (cb->drkey) {
    cb->secqp[n]->modify_to_RTR(HRD_DEFAULT_PSN, remote_qp_attr->qpn,
                                cb->crypto);
  } else {
    cb->secqp[n]->modify_to_RTR(HRD_DEFAULT_PSN, remote_qp_attr->qpn,
                                cb->crypto, key, sizeof(key));
  }

  text(log_fp, "[CLIENT] sequre QP ready to receive\n");
  sleep(1);

  cb->secqp[n]->modify_to_RTS(HRD_DEFAULT_PSN);

  sleep(1);

  return;
}

void hrd_publish_conn_qp(struct hrd_ctrl_blk* cb, int n, const char* qp_name) {
  assert(cb != NULL);
  assert(n >= 0 && n < cb->num_conn_qps);

  assert(qp_name != NULL && strlen(qp_name) < HRD_QP_NAME_SIZE - 1);
  assert(strstr(qp_name, HRD_RESERVED_NAME_PREFIX) == NULL);

  int len = strlen(qp_name);
  int i;
  for (i = 0; i < len; i++) {
    if (qp_name[i] == ' ') {
      fprintf(stderr, "HRD: Space not allowed in QP name\n");
      exit(-1);
    }
  }

  struct hrd_qp_attr qp_attr;
  memcpy(qp_attr.name, qp_name, len);
  qp_attr.name[len] = 0; /* Add the null terminator */
  qp_attr.buf_addr = (uint64_t)(uintptr_t)cb->conn_buf;
  qp_attr.buf_size = cb->conn_buf_size;
  qp_attr.rkey = cb->conn_buf_mr ? cb->conn_buf_mr->rkey : 0;
  qp_attr.qpn = cb->secqp[n]->GetQPN();
  printf("QPN: %u\n", qp_attr.qpn);
  qp_attr.region.length = 0;

  if (cb->withmemkey) {
    // set original region
    ibv_kdf_ctx memkdf =
        initkdf(ibv_kdf_type::KDF_CMAC_AES_128, NULL, KDFKEYSIZE);

    qp_attr.region.begin = (uint64_t)(uintptr_t)cb->conn_buf;
    qp_attr.region.length = cb->conn_buf_size;

    memcpy(qp_attr.region.memkey, mymemkey, KDFKEYSIZE);
    subregion_t subreg =
        grant_subregion(qp_attr.region.begin, 0, qp_attr.region.length,
                        qp_attr.region.length, qp_attr.region.memkey, memkdf);
    assert(subreg.length);

    // set subregion we grant access to
  }

  hrd_publish(qp_attr.name, &qp_attr, sizeof(struct hrd_qp_attr));
}

struct hrd_qp_attr* hrd_get_published_qp(const char* qp_name) {
  struct hrd_qp_attr* ret;
  assert(qp_name != NULL && strlen(qp_name) < HRD_QP_NAME_SIZE - 1);
  assert(strstr(qp_name, HRD_RESERVED_NAME_PREFIX) == NULL);

  int len = strlen(qp_name);
  int i;
  for (i = 0; i < len; i++) {
    if (qp_name[i] == ' ') {
      fprintf(stderr, "HRD: Space not allowed in QP name\n");
      exit(-1);
    }
  }

  int ret_len = hrd_get_published(qp_name, (void**)&ret);

  /*
   * The registry lookup returns only if we get a unique QP for @qp_name, or
   * if the memcached lookup succeeds but we don't have an entry for @qp_name.
   */
  assert(ret_len == sizeof(struct hrd_qp_attr) || ret_len == -1);

  return ret;
}

//#include "../../code/worker/secure_qp.hpp"
/* Every thread creates a TCP connection to the registry only once. */
__thread memcached_st* memc = NULL;

/* Free shm @shm_key and @shm_buf. Return 0 on success, else -1. */
int hrd_free(int shm_key, void* shm_buf) {
  int ret;
  int shmid = shmget(shm_key, 0, 0);
  if (shmid == -1) {
    switch (errno) {
      case EACCES:
        printf(
            "HRD: SHM free error: Insufficient permissions."
            " (SHM key = %d)\n",
            shm_key);
        break;
      case ENOENT:
        printf("HRD: SHM free error: No such SHM key. (SHM key = %d)\n",
               shm_key);
        break;
      default:
        printf("HRD: SHM free error: A wild SHM error: %s\n", strerror(errno));
        break;
    }
    return -1;
  }

  ret = shmctl(shmid, IPC_RMID, NULL);
  if (ret != 0) {
    printf("HRD: SHM free error: shmctl() failed for key %d\n", shm_key);
    exit(-1);
  }

  ret = shmdt(shm_buf);
  if (ret != 0) {
    printf("HRD: SHM free error: shmdt() failed for key %d\n", shm_key);
    exit(-1);
  }

  return 0;
}

/* Like printf, but red. Limited to 1000 characters. */
void hrd_red_printf(const char* format, ...) {
#define RED_LIM 1000
  va_list args;
  int i;

  char buf1[RED_LIM], buf2[RED_LIM];
  memset(buf1, 0, RED_LIM);
  memset(buf2, 0, RED_LIM);

  va_start(args, format);

  /* Marshal the stuff to print in a buffer */
  vsnprintf(buf1, RED_LIM, format, args);

  /* Probably a bad check for buffer overflow */
  for (i = RED_LIM - 1; i >= RED_LIM - 50; i--) {
    assert(buf1[i] == 0);
  }

  /* Add markers for red color and reset color */
  snprintf(buf2, 1000, "\033[31m%s\033[0m", buf1);

  /* Probably another bad check for buffer overflow */
  for (i = RED_LIM - 1; i >= RED_LIM - 50; i--) {
    assert(buf2[i] == 0);
  }

  printf("%s", buf2);

  va_end(args);
}

void hrd_nano_sleep(int ns) {
  long long start = hrd_get_cycles();
  long long end = start;
  int upp = (int)(2.1 * ns);
  while (end - start < upp) {
    end = hrd_get_cycles();
  }
}

/* Return the environment variable @name if it is set. Exit if not. */
char* hrd_getenv(const char* name) {
  char* env = getenv(name);
  if (env == NULL) {
    fprintf(stderr, "Environment variable %s not set\n", name);
    assert(false);
  }

  return env;
}

/* Record the current time in @timebuf. @timebuf must have at least 50 bytes. */
void hrd_get_formatted_time(char* timebuf) {
  assert(timebuf != NULL);
  time_t timer;
  struct tm* tm_info;

  time(&timer);
  tm_info = localtime(&timer);

  strftime(timebuf, 26, "%Y:%m:%d %H:%M:%S", tm_info);
}

memcached_st* hrd_create_memc() {
  memcached_server_st* servers = NULL;
  memcached_st* memc = memcached_create(NULL);
  memcached_return rc;

  memc = memcached_create(NULL);
  char const* registry_ip = "129.132.134.21";  // hrd_getenv("HRD_REGISTRY_IP");

  /* We run the memcached server on the default memcached port */
  servers = memcached_server_list_append(servers, registry_ip,
                                         MEMCACHED_DEFAULT_PORT, &rc);
  rc = memcached_server_push(memc, servers);
  CPE(rc != MEMCACHED_SUCCESS, "Couldn't add memcached server.\n", -1);

  return memc;
}

/*
 * Insert key -> value mapping into memcached running at HRD_REGISTRY_IP.
 */
void hrd_publish(const char* key, void* value, int len) {
  assert(key != NULL && value != NULL && len > 0);
  memcached_return rc;

  if (memc == NULL) {
    memc = hrd_create_memc();
  }

  rc = memcached_set(memc, key, strlen(key), (const char*)value, len, (time_t)0,
                     (uint32_t)0);
  if (rc != MEMCACHED_SUCCESS) {
    char* registry_ip = hrd_getenv("HRD_REGISTRY_IP");
    fprintf(stderr,
            "\tHRD: Failed to publish key %s. Error %s. "
            "Reg IP = %s\n",
            key, memcached_strerror(memc, rc), registry_ip);
    exit(-1);
  }
}

/*
 * Get the value associated with "key" into "value", and return the length
 * of the value. If the key is not found, return NULL and len -1. For all
 * other errors, terminate.
 *
 * This function sometimes gets called in a polling loop - ensure that there
 * are no memory leaks or unterminated memcached connections! We don't need
 * to free() the resul of getenv() since it points to a string in the process
 * environment.
 */
int hrd_get_published(const char* key, void** value) {
  assert(key != NULL);
  if (memc == NULL) {
    memc = hrd_create_memc();
  }

  memcached_return rc;
  size_t value_length;
  uint32_t flags;

  *value = memcached_get(memc, key, strlen(key), &value_length, &flags, &rc);

  if (rc == MEMCACHED_SUCCESS) {
    return (int)value_length;
  } else if (rc == MEMCACHED_NOTFOUND) {
    assert(*value == NULL);
    return -1;
  } else {
    char* registry_ip = hrd_getenv("HRD_REGISTRY_IP");
    fprintf(stderr,
            "HRD: Error finding value for key \"%s\": %s. "
            "Reg IP = %s\n",
            key, memcached_strerror(memc, rc), registry_ip);
    exit(-1);
  }

  /* Never reached */
  assert(false);
}

/*
 * To advertise a queue pair with name qp_name as ready, we publish this
 * key-value mapping: "HRD_RESERVED_NAME_PREFIX-qp_name" -> "hrd_ready". This
 * requires that a qp_name never starts with HRD_RESERVED_NAME_PREFIX.
 *
 * This avoids overwriting the memcached entry for qp_name which might still
 * be needed by the remote peer.
 */
void hrd_publish_ready(const char* qp_name) {
  char value[HRD_QP_NAME_SIZE];
  assert(qp_name != NULL && strlen(qp_name) < HRD_QP_NAME_SIZE);

  char new_name[2 * HRD_QP_NAME_SIZE];
  sprintf(new_name, "%s", HRD_RESERVED_NAME_PREFIX);
  strcat(new_name, qp_name);

  sprintf(value, "%s", "hrd_ready");
  hrd_publish(new_name, value, strlen(value));
}

/*
 * To check if a queue pair with name qp_name is ready, we check if this
 * key-value mapping exists: "HRD_RESERVED_NAME_PREFIX-qp_name" -> "hrd_ready".
 */
void hrd_wait_till_ready(const char* qp_name) {
  char* value;
  char exp_value[HRD_QP_NAME_SIZE];
  sprintf(exp_value, "%s", "hrd_ready");

  char new_name[2 * HRD_QP_NAME_SIZE];
  sprintf(new_name, "%s", HRD_RESERVED_NAME_PREFIX);
  strcat(new_name, qp_name);

  int tries = 0;
  while (true) {
    int ret = hrd_get_published(new_name, (void**)&value);
    tries++;
    if (ret > 0) {
      if (strcmp(value, exp_value) == 0) {
        return;
      }
    }

    usleep(200000);

    if (tries > 100) {
      fprintf(stderr, "HRD: Waiting for QP %s to be ready\n", qp_name);
      tries = 0;
    }
  }
}

#endif /* HRD_H */
